from .helpers import set_optional_arg
from .dynamodb import DynamoDb
from .datastructures import TableDefinition
from .constants import DEFAULT_SELECT


class Table:
    """
        A low-level interface for interacting with DynamoDB Tables
    """
    def __init__(
        self,
        interface: DynamoDb,
        definition: TableDefinition,
        consistent_reads=True,
        read_limit=50
    ):
        self.definition = definition
        self.dynamodb = interface
        self.consistent_reads = consistent_reads
        self.read_limit = read_limit

    @property
    def name(self):
        return self.definition.name

    def query(
        self,
        key_condition_exp: str,
        select: str = DEFAULT_SELECT,
        scan_forward: bool = True,
        start=None,
        index: str = None,
        filter_exp: str = None,
        projection_exp: str = None,
        attr_names: dict = None,
        attr_values: dict = None
    ):
        query_args = {
            "TableName": self.name,
            "KeyConditionExpression": key_condition_exp,
            "Select": select,
            "ScanIndexForward": scan_forward
        }

        set_optional_arg("ExclusiveStartKey", start, query_args)
        set_optional_arg("IndexName", index, query_args)
        set_optional_arg("FilterExpression", filter_exp, query_args)
        set_optional_arg("ExpressionAttributeNames", attr_names, query_args)
        set_optional_arg("ExpressionAttributeValues", attr_values, query_args)
        set_optional_arg("ProjectionExpression", projection_exp, query_args)

        return self.dynamodb.data_client.query(**query_args)

    def scan(
        self,
        select: str = DEFAULT_SELECT,
        scan_forward: bool = True,
        start=None,
        index: str = None,
        filter_exp: str = None,
        projection_exp: str = None,
        attr_names: dict = None,
        attr_values: dict = None
    ):
        scan_args = {
            "TableName": self.name,
            "Select": select,
            "ScanIndexForward": scan_forward
        }

        set_optional_arg("ExclusiveStartKey", start, scan_args)
        set_optional_arg("IndexName", index, scan_args)
        set_optional_arg("FilterExpression", filter_exp, scan_args)
        set_optional_arg("ExpressionAttributeNames", attr_names, scan_args)
        set_optional_arg("ExpressionAttributeValues", attr_values, scan_args)
        set_optional_arg("ProjectionExpression", projection_exp, scan_args)

        return self.dynamodb.data_client.scan(**scan_args)

    def get_item(
        self,
        key: dict,
        projection_exp: str = None,
        attr_names: dict = None
    ):
        get_args = {
            "TableName": self.name,
            "ConsistentRead": self.consistent_reads,
            "Key": key,
        }

        set_optional_arg("ExpressionAttributeNames", attr_names, get_args)
        set_optional_arg("ProjectionExpression", projection_exp, get_args)

        return self.dynamodb.data_client.get_item(**get_args)

    def put_item(
        self,
        item: dict,
        condition_exp: str = None,
        attr_names: dict = None,
        attr_values: dict = None
    ):
        put_args = {
            "TableName": self.name,
            "Item": item
        }

        set_optional_arg("ConditionExpression", condition_exp, put_args)
        set_optional_arg("ExpressionAttributeNames", attr_names, put_args)
        set_optional_arg("ExpressionAttributeValues", attr_values, put_args)

        self.dynamodb.data_client.put_item(**put_args)

    def delete_item(
        self,
        key: dict,
        condition_exp: str = None,
        attr_names: dict = None,
        attr_values: dict = None
    ):
        delete_args = {
            "TableName": self.name,
            "Key": key,
        }

        set_optional_arg("ConditionExpression", condition_exp, delete_args)
        set_optional_arg("ExpressionAttributeNames", attr_names, delete_args)
        set_optional_arg("ExpressionAttributeValues", attr_values, delete_args)

        self.dynamodb.data_client.delete_item(**delete_args)

    def update_item(
        self,
        key: dict,
        update_exp: str,
        condition_exp: str = None,
        attr_names: dict = None,
        attr_values: dict = None
    ):
        update_args = {
            "TableName": self.name,
            "Key": key,
            "UpdateExpression": update_exp,
        }

        set_optional_arg("ConditionExpression", condition_exp, update_args)
        set_optional_arg("ExpressionAttributeNames", attr_names, update_args)
        set_optional_arg("ExpressionAttributeValues", attr_values, update_args)

        self.dynamodb.data_client.update_item(**update_args)

    @classmethod
    def get_table(cls, interface: DynamoDb, name: str):
        return cls(interface, interface.get_table_definition(name))



from dataclasses import field


def default(factory):
    return field(default_factory=factory)


def set_optional_arg(name, val, args):
    if val is not None:
        args[name] = val


def set_arg_if_not_empty(name, val, args):
    if len(val) > 0:
        args[name] = val


def first_kvp(dct: dict):
    return next(iter(dct.items()))


def invert_dict(dct: dict):
    return {v: k for k, v in dct.items()}


def autoclassifier(datatype):
    return lambda x: isinstance(x, datatype)



from dataclasses import dataclass
from typing import Dict, List

from .key_definition import KeyDefinition
from ..helpers import default


@dataclass
class TableDefinition:
    name: str = None
    attributes: Dict[str, str] = default(dict)
    primary_key: KeyDefinition = None
    gsi: Dict[str, KeyDefinition] = default(dict)
    lsi: Dict[str, KeyDefinition] = default(dict)
    arn: str = None
    status: str = None

    @property
    def key_attributes(self) -> List[str]:
        attrs = self.primary_key.attributes

        for key in self.secondary_keys:
            attrs += key.attributes

        return list(set(attrs))

    @property
    def gsi_keys(self) -> List[KeyDefinition]:
        return list(self.gsi.values())

    @property
    def lsi_keys(self) -> List[KeyDefinition]:
        return list(self.lsi.values())

    @property
    def secondary_keys(self) -> List[KeyDefinition]:
        return self.gsi_keys + self.lsi_keys

    @property
    def is_valid(self) -> bool:
        if not self.primary_key.is_valid(self.attributes):
            return False

        for key in self.secondary_keys:
            if not key.is_valid(self.attributes):
                return False
        return True

    @property
    def keys(self) -> List[KeyDefinition]:
        return [self.primary_key] + self.secondary_keys

    def get_secondary_key(self, name):
        if name in self.lsi:
            return self.lsi[name]
        return self.gsi[name]

    def get_projected_attributes(self, index) -> List[str]:
        key = None

        if index in self.gsi:
            key = self.gsi[index]
        else:
            key = self.lsi[index]

        if key.projection == "ALL":
            return list(self.attributes.keys())
        else:
            base_keys = [self.primary_key.partition_key]
            if key.projection == "KEYS_ONLY":
                return base_keys + key.attributes
            elif key.projection == "INCLUDE":
                return base_keys + key.attributes + key.include



from dataclasses import dataclass
from typing import Dict, Any

from ..helpers import default


@dataclass
class Record:
    table: str = None
    from_index: str = None
    attributes: Dict[str, Any] = default(dict)



from numbers import Number

from .marshaller import Marshaller
from base64 import b64encode, b64decode

from ..helpers import autoclassifier
from ..constants import SET_SUBTYPES, ENCODING


class AutoMarshaller(Marshaller):
    """
        This is the default Marshaller implementation.

        This implementation uses datatype checking to infer DynamoDB
        attribute types. This implementation can automatically pack
        appropriate lists into set types if enabled. For example, if
        a value to be packed is a list of strings, this will be stored
        as a String Set.

        Additionally, the casting of DynamoDB Numbers to Python types is
        customisable.
    """

    _DEFAULT_CLASSIFIERS = {
        "S": autoclassifier(str),
        "N": lambda x: (not isinstance(x, bool)) and isinstance(x, Number),
        "B": autoclassifier(bytes),
        "M": autoclassifier(dict),
        "L": autoclassifier(list),
        "NULL": autoclassifier(type(None)),
        "BOOL": autoclassifier(bool)
    }

    def __init__(self, auto_set_detect: bool = True, number_cast=float):
        self.auto_set_detect = auto_set_detect
        self.number_cast = number_cast

    def _detect_default_type(self, value):
        for typekey, classifier in self._DEFAULT_CLASSIFIERS.items():
            if classifier(value):
                return typekey

    def _detect_set_type(self, value: list):
        last_subtype = None

        def is_not_valid_set(subtype):
            is_not_set_type = subtype not in SET_SUBTYPES
            is_not_first = last_subtype is not None
            is_not_match = subtype != last_subtype

            return is_not_set_type or (is_not_first and is_not_match)

        for item in value:
            subtype = self._detect_attribute_type(item, True)
            if is_not_valid_set(subtype):
                return "L"
            last_subtype = subtype
        return f"{last_subtype}S"

    def _detect_attribute_type(self, value, override_set_detect=False):
        should_check_set = not override_set_detect and self.auto_set_detect
        if should_check_set and isinstance(value, list):
            return self._detect_set_type(value)
        return self._detect_default_type(value)

    def detect_attribute_type(self, value):
        return self._detect_attribute_type(value)

    def pack_string(self, value: str):
        return str(value)

    def pack_number(self, value: Number):
        try:
            if int(value) == value:
                value = int(value)
        finally:
            return str(value)

    def pack_binary(self, value: bytes):
        return b64encode(value).decode(ENCODING)

    def pack_bool(self, value: bool):
        return bool(value)

    def unpack_string(self, value):
        return str(value)

    def unpack_number(self, value):
        return self.number_cast(value)

    def unpack_binary(self, value):
        return b64decode(value.encode(ENCODING))

    def unpack_bool(self, value):
        raise bool(value)



from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor


from ..table import Table
from ..constants import DEFAULT_PAGE


class Paginator(ABC):
    def __init__(
        self,
        table: Table,
        base_args: dict,
        page_length: int = DEFAULT_PAGE,
        preload=True
    ):
        self.table = table
        self._items = []
        self._eof = False
        self._executor = ThreadPoolExecutor()
        self._last_page = None
        self._last_key = None
        self._base_args = base_args
        self._page_length = page_length
        if preload:
            self.trigger_page()
            self.ingest_page()

    @property
    def should_page(self):
        return len(self._items) <= self._page_length

    @property
    def can_page(self):
        return self._last_page is None

    @property
    def next_call_args(self):
        if self._last_key is None:
            return self._base_args
        args = {"start": self._last_key}
        args.update(self._base_args)
        return args

    @abstractmethod
    def on_page(self):
        raise NotImplementedError

    def trigger_page(self):
        self._page = self._executor.submit(lambda: self.on_page())

    def ingest_page(self):
        result = self._page.result()
        self._page = None

        if "LastEvaluatedKey" in result:
            self._last_key = result["LastEvaluatedKey"]
        else:
            self._last_key = None
            self._eof = True

        new_items = result["Items"]
        new_items.reverse()

        self._items = new_items + self._items

    def __iter__(self):
        if self._eof:
            raise RuntimeError("Paginator called after EOF")

        while not self._eof and len(self._items) > 0:
            if not self._eof and self.should_page and self.can_page:
                self.trigger_page()

            if self._page.done or len(self._items) == 0:
                self.ingest_page()

            yield self._items.pop()



from abc import ABC, abstractmethod


class RecordMap(ABC):
    @abstractmethod
    def map_record(self, record: dict):
        raise NotImplementedError

    @abstractmethod
    def unmap_object(self, obj) -> dict:
        raise NotImplementedError



from typing import Dict
from dataclasses import dataclass

from .datastructures import KeyDefinition
from .table import Table
from .marshallers import Marshaller, AutoMarshaller
from .recordmaps import RecordMap
from .paginators import QueryPaginator, ScanPaginator, Paginator  # noqa
from .constants import DEFAULT_SELECT  # noqa


class TableView:
    """
        A high-level interface for DynamoDB Tables
        Supports the following features:
            - Automatic expression synthesis for known keys
            - Attribute marshalling/unmarshalling
            - Datatype and class mapping
            - Result pagination
    """

    @dataclass
    class SecondaryIndex:
        name: str
        definition: KeyDefinition
        view: "TableView"

        def query(self):
            pass

    def __init__(
        self,
        table: Table,
        marshaller: Marshaller = None,
        recordmap: RecordMap = None
    ):
        self.table = table
        self.marshaller = marshaller
        self.recordmap = recordmap
        self.secondary_indexes = []

    def _make_index(self, name, definition):
        return self.SecondaryIndex(name, definition, self)

    def _build_secondary_indexes(self):
        def _import_from_dict(dct: Dict[str, KeyDefinition]):
            for k, v in dct.items():
                self.secondary_indexes.append(self._make_index(k, v))
        _import_from_dict(self.table.definition.lsi)
        _import_from_dict(self.table.definition.gsi)

    def _postprocess_read_record(self, record: dict):
        if self.marshaller is not None:
            record = {
                k: self.marshaller.unpack_attribute(v)
                for k, v
                in record.items()
            }
        if self.recordmap is not None:
            return self.recordmap.map_record(record)
        return record

    def _preprocess_write_record(self, data):
        if self.recordmap is not None:
            data = self.recordmap.unmap_object(data)
        if self.marshaller is not None:
            return {
                k: self.marshaller.pack_attribute(v) for k, v in data.items()
            }
        return data

    @classmethod
    def get_default_table_view(cls, table: Table):
        return cls(table, AutoMarshaller())

    def query(self, key_condition_exp, **kwargs) -> Paginator:
        kwargs["key_condition_exp"] = key_condition_exp

    def _get_primary_key_selector(self, value, sort=False):
        primary_key = self.table.definition.primary_key
        key_name = primary_key.partition_key

        if sort:
            assert primary_key.is_sortable
            key_name = primary_key.sort_key

        key_type = self.table.definition.attributes[key_name]

        return {key_name: {key_type: value}}

    def _build_key_args(self, partition_key, sort_key):
        key_args = self._get_primary_key_selector(partition_key)
        if sort_key is not None:
            key_args.update(self._get_primary_key_selector(sort_key, True))
        return key_args



import time
from typing import List

from boto3 import Session
from amazondax import AmazonDaxClient
from boto3_type_annotations import dynamodb

from .helpers import set_arg_if_not_empty
from .datastructures import TableDefinition, KeyDefinition


class DynamoDb:
    def __init__(self, session: Session, dax_endpoints: List[str] = None):
        self.session = session
        self.client = session.client("dynamodb")
        self.dax = None

        if dax_endpoints is not None:
            self.dax = AmazonDaxClient(session, endpoints=dax_endpoints)

    def list_tables(self) -> List[str]:
        return self.client .list_tables()["TableNames"]

    def get_table_definition(self, name):
        definition = TableDefinition()
        response = self.client.describe_table(TableName=name)
        table = response["Table"]

        definition.name = table["TableName"]
        definition.arn = table["TableArn"]
        definition.status = table["TableStatus"]

        if "AttributeDefinitions" in table:
            for attr in table["AttributeDefinitions"]:
                attr_name = attr["AttributeName"]
                definition.attributes[attr_name] = attr["AttributeType"]

        if "KeySchema" in table:
            definition.primary_key = KeyDefinition.from_key_schema(
                table["KeySchema"]
            )

        indexes = (("Global", definition.gsi), ("Local", definition.lsi))

        for index_scope, index_dict in indexes:
            index_type = f"{index_scope}SecondaryIndexes"
            if index_type in table:
                for index in table[index_type]:
                    keydef = KeyDefinition.from_key_schema(index["KeySchema"])
                    index_dict[index["IndexName"]] = keydef

        return definition

    def create_table(self, definition: TableDefinition, **kwargs):
        creation_args = self.get_creation_args_for_definition(definition)
        creation_args.update(kwargs)
        self.client.create_table(**creation_args)
        return self.get_table_definition(definition.name)

    def delete_table(self, table):
        table = self._normalise_table_name(table)
        self.client.delete_table(TableName=table)

    def wait_for_states(self, table, states):
        table = self._normalise_table_name(table)
        ready = False

        while not ready:
            tabledef = self.get_table_definition(table)
            ready = tabledef.status in states
            time.sleep(1)

    def wait_for_active(self, table):
        return self.wait_for_states(table, ["ACTIVE"])

    def wait_for_deletion(self, table):
        table = self._normalise_table_name(table)
        ready = False

        while not ready:
            try:
                self.get_table_definition(table)
                time.sleep(1)
            except self._not_found:
                ready = True

    @property
    def _not_found(self):
        return self.client.exceptions.ResourceNotFoundException

    @staticmethod
    def _normalise_table_name(table):
        if isinstance(table, TableDefinition):
            table = table.name
        return table

    @staticmethod
    def get_creation_args_for_definition(definition: TableDefinition):
        def get_attr_args(key_attribute: str):
            return {
                "AttributeName": key_attribute,
                "AttributeType": definition.attributes[key_attribute]
            }

        def get_index_args(name: str, keydef: KeyDefinition):
            return {
                "IndexName": name,
                "KeySchema": keydef.to_key_schema(),
                "Projection": keydef.to_projection_schema()
            }

        attrs = [get_attr_args(item) for item in definition.key_attributes]

        creation_args = {
            "AttributeDefinitions": attrs,
            "TableName": definition.name,
            "KeySchema": definition.primary_key.to_key_schema(),
            "BillingMode": "PAY_PER_REQUEST"
        }

        lsi = [
            get_index_args(key, value)
            for key, value
            in definition.lsi.items()
        ]

        gsi = [
            get_index_args(key, value)
            for key, value
            in definition.gsi.items()
        ]

        set_arg_if_not_empty("LocalSecondaryIndexes", lsi, creation_args)
        set_arg_if_not_empty("GlobalSecondaryIndexes", gsi, creation_args)

        return creation_args

    @classmethod
    def get_default_client(cls, dax_endpoints=None):
        return cls(Session(), dax_endpoints)

    @property
    def data_client(self) -> dynamodb.Client:
        if self.dax is not None:
            return self.dax
        return self.client



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="byrne",
    version="0.1.0",
    author="Jacob Neil Taylor",
    author_email="me@jacobtaylor.id.au",
    description="A intelligent DynamoDB frontend for Python3",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jacobneiltaylor/byrne",
    package_dir={'': 'src'},
    packages=setuptools.find_packages(where='src'),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



from dataclasses import dataclass
from typing import Dict, List
from ..helpers import default


_KEY_DATATYPES = [
    "S",
    "N",
    "B"
]

_PARTITION_KEY = "HASH"
_SORT_KEY = "RANGE"


@dataclass
class KeyDefinition:
    partition_key: str
    sort_key: str = None
    projection: str = "ALL"
    include: List[str] = default(list)

    @property
    def is_sortable(self):
        if not self.sort_key:
            return False
        return True

    @property
    def attributes(self) -> List[str]:
        if self.is_sortable:
            return [self.partition_key, self.sort_key]
        return [self.partition_key]

    def is_valid(self, attributes: Dict[str, str]) -> bool:
        def is_key_valid(key):
            return key in attributes and attributes[key] in _KEY_DATATYPES
        if self.is_sortable:
            return is_key_valid(self.partition_key) and is_key_valid(self.sort_key)  # noqa: E501
        return is_key_valid(self.partition_key)

    def to_key_schema(self):
        def get_schema_item(name, partition=True):
            return {
                "AttributeName": name,
                "KeyType": "HASH" if partition else "RANGE",
            }
        schema = [get_schema_item(self.partition_key)]
        if self.is_sortable:
            schema.append(get_schema_item(self.sort_key, False))
        return schema

    def to_projection_schema(self):
        schema = {
            "ProjectionType": self.projection,
        }

        if self.projection == "INCLUDE":
            schema["NonKeyAttributes"] = self.include

        return schema

    @classmethod
    def from_key_schema(cls, schema: List[Dict[str, str]]):
        partition_key = None
        sort_key = None

        for key in schema:
            key_type = key["KeyType"]
            if key_type == _PARTITION_KEY:
                partition_key = key["AttributeName"]
            elif key_type == _SORT_KEY:
                sort_key = key["AttributeName"]

        assert(partition_key is not None)

        return cls(partition_key, sort_key)



from abc import ABC, abstractmethod
from numbers import Number

from ..helpers import first_kvp


class Marshaller(ABC):
    """
        A Marshaller is responsible for translating between
        the "packed" representation of attribute values from DynamoDB
        and a "native" Python representation of that value and vice versa.

        For example, consider the following attribute value from DDB:
            {'BS': ['MQ==', 'Mg==', 'Mw==']}
        This is represented in Python as:
            [b"1", b"2", b"3"]
    """
    @abstractmethod
    def detect_attribute_type(self, value) -> str:
        raise NotImplementedError

    @abstractmethod
    def pack_string(self, value: str):
        raise NotImplementedError

    @abstractmethod
    def pack_number(self, value: Number):
        raise NotImplementedError

    @abstractmethod
    def pack_binary(self, value: bytes):
        raise NotImplementedError

    @abstractmethod
    def pack_bool(self, value: bool):
        raise NotImplementedError

    def pack_null(self, value: type(None)):
        return True

    def pack_string_set(self, value: list):
        return [self.pack_string(x) for x in value]

    def pack_number_set(self, value: list):
        return [self.pack_number(x) for x in value]

    def pack_binary_set(self, value: list):
        return [self.pack_binary(x) for x in value]

    def pack_map(self, value: dict):
        return {k: self.pack_attribute(v) for k, v in value.items()}

    def pack_list(self, value: list):
        return [self.pack_attribute(i) for i in value]

    @abstractmethod
    def unpack_string(self, value):
        raise NotImplementedError

    @abstractmethod
    def unpack_number(self, value):
        raise NotImplementedError

    @abstractmethod
    def unpack_binary(self, value):
        raise NotImplementedError

    @abstractmethod
    def unpack_bool(self, value):
        raise NotImplementedError

    def unpack_null(self, value):
        return None

    def unpack_string_set(self, value: list):
        return [self.unpack_string(x) for x in value]

    def unpack_number_set(self, value: list):
        return [self.unpack_number(x) for x in value]

    def unpack_binary_set(self, value: list):
        return [self.unpack_binary(x) for x in value]

    def unpack_map(self, data):
        return {k: self.unpack_attribute(v) for k, v in data.items()}

    def unpack_list(self, data):
        return [self.unpack_attribute(i) for i in data]

    def unpack_attribute(self, attribute):
        attribute_type, attribute_value = first_kvp(attribute)

        return {
            "S": self.unpack_string,
            "N": self.unpack_number,
            "B": self.unpack_binary,
            "SS": self.unpack_string_set,
            "NS": self.unpack_number_set,
            "BS": self.unpack_binary_set,
            "M": self.unpack_map,
            "L": self.unpack_list,
            "NULL": self.unpack_null,
            "BOOL": self.unpack_bool
        }[attribute_type](attribute_value)

    def pack_attribute(self, value):
        attribute_type = self.detect_attribute_type(value)
        attribute_value = {
            "S": self.pack_string,
            "N": self.pack_number,
            "B": self.pack_binary,
            "SS": self.pack_string_set,
            "NS": self.pack_number_set,
            "BS": self.pack_binary_set,
            "M": self.pack_map,
            "L": self.pack_list,
            "NULL": self.pack_null,
            "BOOL": self.pack_bool
        }[attribute_type](value)
        return {attribute_type: attribute_value}
